@@ -603,6 +603,10 @@ class RequestConfiguration:
 
         return instance
 
+    @classmethod
+    def set_instance(cls, instance: 'RequestConfiguration') -> None:
+        _request_info.set(instance)  # type: ignore
+
     @classmethod
     def clear(cls):
         """
